@@ -73,12 +73,24 @@ void deleteRecursively(const QString &dirName)
     QDir().rmdir(dirName);
 }
 
+FILE *openProcess(const QString &command)
+{
+#if defined(Q_OS_WIN32)
+    QString processedCommand = QLatin1Char('\"') + command + QLatin1Char('\"');
+#else
+    QString processedCommand = command;
+#endif
+
+    return popen(processedCommand.toLocal8Bit().constData(), "r");
+}
+
 struct Options
 {
     Options()
         : helpRequested(false)
         , verbose(false)
         , timing(false)
+        , generateAssetsFileList(true)
         , minimumAndroidVersion(9)
         , targetAndroidVersion(10)
         , deploymentMechanism(Bundled)
@@ -108,6 +120,7 @@ struct Options
     bool helpRequested;
     bool verbose;
     bool timing;
+    bool generateAssetsFileList;
     QTime timer;
 
     // External tools
@@ -300,6 +313,8 @@ Options parseOptions()
             options.sectionsOnly = true;
         } else if (argument.compare(QLatin1String("--protected"), Qt::CaseInsensitive) == 0) {
             options.protectedAuthenticationPath = true;
+        } else if (argument.compare(QLatin1String("--no-generated-assets-cache"), Qt::CaseInsensitive) == 0) {
+            options.generateAssetsFileList = false;
         }
     }
 
@@ -333,7 +348,7 @@ void printHelp()
                     "       platform. By default, the highest available version will be\n"
                     "       used.\n"
                     "    --ant <path/to/ant>: If unspecified, ant from the PATH will be\n"
-                    "       used."
+                    "       used.\n"
                     "    --release: Builds a package ready for release. By default, the\n"
                     "       package will be signed with a debug key.\n"
                     "    --sign <url/to/keystore> <alias>: Signs the package with the\n"
@@ -359,6 +374,8 @@ void printHelp()
                     "       an attempt is made to detect the tool using the JAVA_HOME and\n"
                     "       PATH environment variables, in that order.\n"
                     "    --verbose: Prints out information during processing.\n"
+                    "    --no-generated-assets-cache: Do not pregenerate the entry list for\n"
+                    "       the assets file engine.\n"
                     "    --help: Displays this information.\n\n",
                     qPrintable(QCoreApplication::arguments().at(0))
             );
@@ -388,13 +405,18 @@ bool alwaysOverwritableFile(const QString &fileName)
             || fileName.endsWith(QLatin1String("/src/org/qtproject/qt5/android/bindings/QtActivity.java")));
 }
 
-bool copyFileIfNewer(const QString &sourceFileName, const QString &destinationFileName, bool verbose)
+bool copyFileIfNewer(const QString &sourceFileName,
+                     const QString &destinationFileName,
+                     bool verbose,
+                     bool forceOverwrite = false)
 {
     if (QFile::exists(destinationFileName)) {
         QFileInfo destinationFileInfo(destinationFileName);
         QFileInfo sourceFileInfo(sourceFileName);
 
-        if (sourceFileInfo.lastModified() <= destinationFileInfo.lastModified() && !alwaysOverwritableFile(destinationFileName)) {
+        if (!forceOverwrite
+                && sourceFileInfo.lastModified() <= destinationFileInfo.lastModified()
+                && !alwaysOverwritableFile(destinationFileName)) {
             if (verbose)
                 fprintf(stdout, "  -- Skipping file %s. Same or newer file already in place.\n", qPrintable(sourceFileName));
             return true;
@@ -460,8 +482,9 @@ QString cleanPackageName(QString packageName)
         QString word = packageName.mid(index + 1, next - index - 1);
         if (!word.isEmpty()) {
             QChar c = word[0];
-            if (c >= QChar(QLatin1Char('0')) && c<= QChar(QLatin1Char('9'))) {
-                packageName.insert(index + 1, QLatin1Char('_'));
+            if ((c >= QChar(QLatin1Char('0')) && c<= QChar(QLatin1Char('9')))
+                   || c == QLatin1Char('_')) {
+                packageName.insert(index + 1, QLatin1Char('a'));
                 index = next + 1;
                 continue;
             }
@@ -636,7 +659,7 @@ bool readInputFile(Options *options)
     return true;
 }
 
-bool copyFiles(const QDir &sourceDirectory, const QDir &destinationDirectory, bool verbose)
+bool copyFiles(const QDir &sourceDirectory, const QDir &destinationDirectory, bool verbose, bool forceOverwrite = false)
 {
     QFileInfoList entries = sourceDirectory.entryInfoList(QDir::NoDotAndDotDot | QDir::Files | QDir::Dirs);
     foreach (QFileInfo entry, entries) {
@@ -647,11 +670,11 @@ bool copyFiles(const QDir &sourceDirectory, const QDir &destinationDirectory, bo
                 return false;
             }
 
-            if (!copyFiles(dir, QDir(destinationDirectory.path() + QLatin1String("/") + dir.dirName()), verbose))
+            if (!copyFiles(dir, QDir(destinationDirectory.path() + QLatin1String("/") + dir.dirName()), verbose, forceOverwrite))
                 return false;
         } else {
             QString destination = destinationDirectory.absoluteFilePath(entry.fileName());
-            if (!copyFileIfNewer(entry.absoluteFilePath(), destination, verbose))
+            if (!copyFileIfNewer(entry.absoluteFilePath(), destination, verbose, forceOverwrite))
                 return false;
         }
     }
@@ -692,7 +715,7 @@ bool copyAndroidSources(const Options &options)
         return false;
     }
 
-    return copyFiles(sourceDirectory, QDir(options.outputDirectory), options.verbose);
+    return copyFiles(sourceDirectory, QDir(options.outputDirectory), options.verbose, true);
 }
 
 bool copyAndroidExtraLibs(const Options &options)
@@ -857,8 +880,9 @@ bool updateAndroidManifest(Options &options)
     }
 
     bool usesGL = false;
-    foreach (QString localLib, localLibs) {
-        if (localLib.endsWith(QLatin1String("libqtforandroidGL.so"))) {
+    foreach (QString qtDependency, options.qtDependencies) {
+        if (qtDependency.endsWith(QLatin1String("libQt5OpenGL.so"))
+                || qtDependency.endsWith(QLatin1String("libQt5Quick.so"))) {
             usesGL = true;
             break;
         }
@@ -877,14 +901,14 @@ bool updateAndroidManifest(Options &options)
 
     QString permissions;
     foreach (QString permission, options.permissions)
-        permissions += QString::fromLatin1("<uses-permission android:name=\"%1\" />\n").arg(permission);
+        permissions += QString::fromLatin1("    <uses-permission android:name=\"%1\" />\n").arg(permission);
     replacements[QLatin1String("<!-- %%INSERT_PERMISSIONS -->")] = permissions;
 
     QString features;
     foreach (QString feature, options.features)
-        features += QString::fromLatin1("<uses-feature android:name=\"%1\" />\n").arg(feature);
+        features += QStringLiteral("    <uses-feature android:name=\"%1\" android:required=\"false\" />\n").arg(feature);
     if (usesGL)
-        features += QStringLiteral("<uses-feature android:glEsVersion=\"0x00020000\" android:required=\"true\" />");
+        features += QStringLiteral("    <uses-feature android:glEsVersion=\"0x00020000\" android:required=\"true\" />");
 
     replacements[QLatin1String("<!-- %%INSERT_FEATURES -->")] = features;
 
@@ -1174,7 +1198,7 @@ QStringList getQtLibsFromElf(const Options &options, const QString &fileName)
 
     readElf = QString::fromLatin1("\"%1\" -d -W %2").arg(readElf).arg(fileName);
 
-    FILE *readElfCommand = popen(readElf.toLocal8Bit().constData(), "r");
+    FILE *readElfCommand = openProcess(readElf);
     if (readElfCommand == 0) {
         fprintf(stderr, "Cannot execute command %s", qPrintable(readElf));
         return QStringList();
@@ -1263,7 +1287,7 @@ bool readDependencies(Options *options)
 
     while (!remainingDependencies.isEmpty()) {
         QSet<QString>::iterator start = remainingDependencies.begin();
-        QString fileName = qtDir+*start;
+        QString fileName = qtDir + *start;
         remainingDependencies.erase(start);
 
         QStringList unmetDependencies;
@@ -1277,6 +1301,21 @@ bool readDependencies(Options *options)
                     qPrintable(unmetDependencies.join(QLatin1Char(','))));
         }
     }
+
+    QStringList::iterator it = options->localLibs.begin();
+    while (it != options->localLibs.end()) {
+        QStringList unmetDependencies;
+        if (!goodToCopy(options, qtDir + *it, &unmetDependencies)) {
+            if (options->verbose) {
+                fprintf(stdout, "Skipping %s due to unmet dependencies: %s\n",
+                        qPrintable(*it),
+                        qPrintable(unmetDependencies.join(QLatin1Char(','))));
+            }
+            it = options->localLibs.erase(it);
+        } else {
+            ++it;
+        }
+    }
     return true;
 }
 
@@ -1303,7 +1342,7 @@ bool stripFile(const Options &options, const QString &fileName)
 
     strip = QString::fromLatin1("\"%1\" %2").arg(strip).arg(fileName);
 
-    FILE *stripCommand = popen(strip.toLocal8Bit().constData(), "r");
+    FILE *stripCommand = openProcess(strip);
     if (stripCommand == 0) {
         fprintf(stderr, "Cannot execute command %s", qPrintable(strip));
         return false;
@@ -1384,7 +1423,7 @@ FILE *runAdb(const Options &options, const QString &arguments)
     if (options.verbose)
         fprintf(stdout, "Running command \"%s\"\n", adb.toLocal8Bit().constData());
 
-    FILE *adbCommand = popen(adb.toLocal8Bit().constData(), "r");
+    FILE *adbCommand = openProcess(adb);
     if (adbCommand == 0) {
         fprintf(stderr, "Cannot start adb: %s\n", qPrintable(adb));
         return 0;
@@ -1520,6 +1559,8 @@ bool copyQtFiles(Options *options)
         // For debug deployment, we copy all libraries and plugins
         QDirIterator dirIterator(options->qtInstallDirectory, QDirIterator::Subdirectories);
         while (dirIterator.hasNext()) {
+            dirIterator.next();
+
             QFileInfo info = dirIterator.fileInfo();
             if (!info.isDir()) {
                 QString relativePath = info.absoluteFilePath().mid(options->qtInstallDirectory.length());
@@ -1535,7 +1576,6 @@ bool copyQtFiles(Options *options)
                         return false;
                 }
             }
-            dirIterator.next();
         }
 
         foreach (QString qtDependency, options->qtDependencies)
@@ -1619,7 +1659,7 @@ bool createAndroidProject(const Options &options)
     if (options.verbose)
         fprintf(stdout, "  -- Command: %s\n", qPrintable(androidTool));
 
-    FILE *androidToolCommand = popen(androidTool.toLocal8Bit().constData(), "r");
+    FILE *androidToolCommand = openProcess(androidTool);
     if (androidToolCommand == 0) {
         fprintf(stderr, "Cannot run command '%s'\n", qPrintable(androidTool));
         return false;
@@ -1679,7 +1719,7 @@ bool buildAndroidProject(const Options &options)
 
     QString ant = QString::fromLatin1("\"%1\" %2").arg(antTool).arg(options.releasePackage ? QLatin1String(" release") : QLatin1String(" debug"));
 
-    FILE *antCommand = popen(ant.toLocal8Bit().constData(), "r");
+    FILE *antCommand = openProcess(ant);
     if (antCommand == 0) {
         fprintf(stderr, "Cannot run ant command: %s\n.", qPrintable(ant));
         return false;
@@ -1836,29 +1876,29 @@ bool signPackage(const Options &options)
         return false;
     }
 
-    jarSignerTool = QString::fromLatin1("\"%1\" -sigalg %2 -digestalg %3 -keystore %4")
+    jarSignerTool = QString::fromLatin1("\"%1\" -sigalg \"%2\" -digestalg \"%3\" -keystore \"%4\"")
             .arg(jarSignerTool).arg(options.sigAlg).arg(options.digestAlg).arg(options.keyStore);
 
     if (!options.keyStorePassword.isEmpty())
-        jarSignerTool += QString::fromLatin1(" -storepass %1").arg(options.keyStorePassword);
+        jarSignerTool += QString::fromLatin1(" -storepass \"%1\"").arg(options.keyStorePassword);
 
     if (!options.storeType.isEmpty())
-        jarSignerTool += QString::fromLatin1(" -storetype %1").arg(options.storeType);
+        jarSignerTool += QString::fromLatin1(" -storetype \"%1\"").arg(options.storeType);
 
     if (!options.keyPass.isEmpty())
-        jarSignerTool += QString::fromLatin1(" -keypass %1").arg(options.keyPass);
+        jarSignerTool += QString::fromLatin1(" -keypass \"%1\"").arg(options.keyPass);
 
     if (!options.sigFile.isEmpty())
-        jarSignerTool += QString::fromLatin1(" -sigfile %1").arg(options.sigFile);
+        jarSignerTool += QString::fromLatin1(" -sigfile \"%1\"").arg(options.sigFile);
 
     if (!options.signedJar.isEmpty())
-        jarSignerTool += QString::fromLatin1(" -signedjar %1").arg(options.signedJar);
+        jarSignerTool += QString::fromLatin1(" -signedjar \"%1\"").arg(options.signedJar);
 
     if (!options.tsaUrl.isEmpty())
-        jarSignerTool += QString::fromLatin1(" -tsa %1").arg(options.tsaUrl);
+        jarSignerTool += QString::fromLatin1(" -tsa \"%1\"").arg(options.tsaUrl);
 
     if (!options.tsaCert.isEmpty())
-        jarSignerTool += QString::fromLatin1(" -tsacert %1").arg(options.tsaCert);
+        jarSignerTool += QString::fromLatin1(" -tsacert \"%1\"").arg(options.tsaCert);
 
     if (options.internalSf)
         jarSignerTool += QLatin1String(" -internalsf");
@@ -1869,14 +1909,14 @@ bool signPackage(const Options &options)
     if (options.protectedAuthenticationPath)
         jarSignerTool += QLatin1String(" -protected");
 
-    jarSignerTool += QString::fromLatin1(" %1 %2")
+    jarSignerTool += QString::fromLatin1(" %1 \"%2\"")
             .arg(options.outputDirectory
                  + QLatin1String("/bin/")
                  + apkName(options)
                  + QLatin1String("-unsigned.apk"))
             .arg(options.keyStoreAlias);
 
-    FILE *jarSignerCommand = popen(jarSignerTool.toLocal8Bit().constData(), "r");
+    FILE *jarSignerCommand = openProcess(jarSignerTool);
     if (jarSignerCommand == 0) {
         fprintf(stderr, "Couldn't run jarsigner.\n");
         return false;
@@ -1918,7 +1958,7 @@ bool signPackage(const Options &options)
                  + apkName(options)
                  + QLatin1String(".apk"));
 
-    FILE *zipAlignCommand = popen(zipAlignTool.toLocal8Bit(), "r");
+    FILE *zipAlignCommand = openProcess(zipAlignTool);
     if (zipAlignCommand == 0) {
         fprintf(stderr, "Couldn't run zipalign.\n");
         return false;
@@ -1975,7 +2015,7 @@ bool deployAllToLocalTmp(const Options &options)
 {
     FILE *adbCommand = runAdb(options,
                               QString::fromLatin1(" push %1 /data/local/tmp/qt/")
-                              .arg(options.temporaryDirectoryName + QLatin1Char('/')));
+                              .arg(options.temporaryDirectoryName));
     if (adbCommand == 0)
         return false;
 
@@ -1995,6 +2035,58 @@ bool deployAllToLocalTmp(const Options &options)
     return true;
 }
 
+bool generateAssetsFileList(const Options &options)
+{
+    if (options.verbose)
+        fprintf(stdout, "Pregenerating entry list for assets file engine.\n");
+
+    QString assetsPath = options.outputDirectory + QLatin1String("/assets/");
+    QString addedByAndroidDeployQtPath = assetsPath + QLatin1String("--Added-by-androiddeployqt--/");
+    if (!QDir().mkpath(addedByAndroidDeployQtPath)) {
+        fprintf(stderr, "Failed to create directory '%s'", qPrintable(addedByAndroidDeployQtPath));
+        return false;
+    }
+
+    QFile file(addedByAndroidDeployQtPath + QLatin1String("/qt_cache_pregenerated_file_list"));
+    if (file.open(QIODevice::WriteOnly)) {
+        QDirIterator dirIterator(assetsPath,
+                                 QDir::Dirs | QDir::Files | QDir::NoDotAndDotDot,
+                                 QDirIterator::Subdirectories);
+
+        QHash<QString, QStringList> directoryContents;
+        while (dirIterator.hasNext()) {
+            const QString name = dirIterator.next().mid(assetsPath.length());
+
+            int slashIndex = name.lastIndexOf(QLatin1Char('/'));
+            QString pathName = slashIndex >= 0 ? name.left(slashIndex) : QString::fromLatin1("/");
+            QString fileName = slashIndex >= 0 ? name.mid(pathName.length() + 1) : name;
+
+            if (!fileName.isEmpty() && dirIterator.fileInfo().isDir() && !fileName.endsWith(QLatin1Char('/')))
+                fileName += QLatin1Char('/');
+
+            if (fileName.isEmpty() && !directoryContents.contains(pathName))
+                directoryContents[pathName] = QStringList();
+            else if (!fileName.isEmpty())
+                directoryContents[pathName].append(fileName);
+        }
+
+        QDataStream stream(&file);
+        stream.setVersion(QDataStream::Qt_5_3);
+        QList<QString> directories = directoryContents.keys();
+        foreach (const QString &directory, directories) {
+            QStringList entryList = directoryContents.value(directory);
+            stream << directory << entryList.size();
+            foreach (const QString &entry, entryList)
+                stream << entry;
+        }
+    } else {
+        fprintf(stderr, "Pregenerating entry list for assets file engine failed!\n");
+        return false;
+    }
+
+    return true;
+}
+
 enum ErrorCode
 {
     Success,
@@ -2014,7 +2106,8 @@ enum ErrorCode
     CannotBuildAndroidProject = 14,
     CannotSignPackage = 15,
     CannotInstallApk = 16,
-    CannotDeployAllToLocalTmp = 17
+    CannotDeployAllToLocalTmp = 17,
+    CannotGenerateAssetsFileList = 18
 };
 
 int main(int argc, char *argv[])
@@ -2113,6 +2206,9 @@ int main(int argc, char *argv[])
     if (!updateAndroidFiles(options))
         return CannotUpdateAndroidFiles;
 
+    if (options.generateAssetsFileList && !generateAssetsFileList(options))
+        return CannotGenerateAssetsFileList;
+
     if (Q_UNLIKELY(options.timing))
         fprintf(stdout, "[TIMING] %d ms: Updated files\n", options.timer.elapsed());
 
