@@ -213,11 +213,11 @@ CentralWidget::CentralWidget(QWidget *parent)
 
     connect(m_findWidget, SIGNAL(findNext()), this, SLOT(findNext()));
     connect(m_findWidget, SIGNAL(findPrevious()), this, SLOT(findPrevious()));
-    connect(m_findWidget, SIGNAL(find(QString, bool, bool)), this,
-        SLOT(find(QString, bool, bool)));
+    connect(m_findWidget, SIGNAL(find(QString,bool,bool)), this,
+        SLOT(find(QString,bool,bool)));
     connect(m_findWidget, SIGNAL(escapePressed()), this, SLOT(activateTab()));
-    connect(m_tabBar, SIGNAL(addBookmark(QString, QString)), this,
-        SIGNAL(addBookmark(QString, QString)));
+    connect(m_tabBar, SIGNAL(addBookmark(QString,QString)), this,
+        SIGNAL(addBookmark(QString,QString)));
 }
 
 CentralWidget::~CentralWidget()
@@ -616,6 +616,8 @@ void CentralWidget::connectSignals(HelpViewer *page)
 {
     TRACE_OBJ
 #if !defined(QT_NO_WEBKIT)
+    connect(page, SIGNAL(printRequested()), this, SLOT(print()));
+#endif
     connect(page, SIGNAL(copyAvailable(bool)), this,
         SIGNAL(copyAvailable(bool)));
     connect(page, SIGNAL(forwardAvailable(bool)), this,
@@ -624,11 +626,7 @@ void CentralWidget::connectSignals(HelpViewer *page)
         SIGNAL(backwardAvailable(bool)));
     connect(page, SIGNAL(sourceChanged(QUrl)), this,
         SLOT(handleSourceChanged(QUrl)));
-    connect(page, SIGNAL(printRequested()), this, SLOT(print()));
     connect(page, SIGNAL(highlighted(QString)), this, SLOT(slotHighlighted(QString)));
-#else
-    Q_UNUSED(page)
-#endif
 }
 
 bool CentralWidget::eventFilter(QObject *object, QEvent *e)
