@@ -87,6 +87,11 @@ public:
     void setEditorFocus(int model);
     void setUnfinishedEditorFocus();
     bool focusNextUnfinished();
+    void setVisualizeWhitespace(bool value);
+    void setFontSize(const float fontSize);
+    float fontSize();
+    void increaseFontSize();
+    void decreaseFontSize();
 
 signals:
     void translationChanged(const QStringList &translations);
@@ -157,6 +162,7 @@ private:
     void addPluralForm(int model, const QString &label, bool writable);
     void fixTabOrder();
     QPalette paletteForModel(int model) const;
+    void applyFontSize();
 
     MultiDataModel *m_dataModel;
 
@@ -165,6 +171,7 @@ private:
     int m_currentNumerus;
 
     bool m_lengthVariants;
+    float m_fontSize;
 
     bool m_undoAvail;
     bool m_redoAvail;
@@ -172,6 +179,7 @@ private:
     bool m_copyAvail;
 
     bool m_clipboardEmpty;
+    bool m_visualizeWhitespace;
 
     QTextEdit *m_selectionHolder;
     QWidget *m_focusWidget;
