@@ -1,6 +1,6 @@
 /****************************************************************************
 **
-** Copyright (C) 2013 Digia Plc and/or its subsidiary(-ies).
+** Copyright (C) 2014 Digia Plc and/or its subsidiary(-ies).
 ** Contact: http://www.qt-project.org/legal
 **
 ** This file is part of the tools applications of the Qt Toolkit.
@@ -52,6 +52,7 @@
 #if defined(Q_OS_WIN)
 #  include <QtCore/qt_windows.h>
 #  include <Shlwapi.h>
+#  include <delayimp.h>
 #else // Q_OS_WIN
 #  include <sys/wait.h>
 #  include <sys/types.h>
@@ -63,8 +64,6 @@
 #  include <fcntl.h>
 #endif  // !Q_OS_WIN
 
-#include <cstdio>
-
 QT_BEGIN_NAMESPACE
 
 int optVerboseLevel = 1;
@@ -106,7 +105,7 @@ bool createDirectory(const QString &directory, QString *errorMessage)
         return false;
     }
     if (optVerboseLevel)
-        std::printf("Creating %s...\n", qPrintable(QDir::toNativeSeparators(directory)));
+        std::wcout << "Creating " << QDir::toNativeSeparators(directory) << "...\n";
     QDir dir;
     if (!dir.mkpath(directory)) {
         *errorMessage = QString::fromLatin1("Could not create directory %1.").
@@ -132,11 +131,12 @@ QStringList findSharedLibraries(const QDir &directory, Platform platform, bool d
         bool matches = true;
         if (platform & WindowsBased) {
             bool debugDll;
-            if (readPeExecutable(dllPath, &errorMessage, 0, 0, &debugDll)) {
+            if (readPeExecutable(dllPath, &errorMessage, 0, 0, &debugDll,
+                                 (platform == WindowsMinGW))) {
                 matches = debugDll == debug;
             } else {
-                std::fprintf(stderr, "Warning: Unable to read %s: %s",
-                             qPrintable(QDir::toNativeSeparators(dllPath)), qPrintable(errorMessage));
+                std::wcerr << "Warning: Unable to read " << QDir::toNativeSeparators(dllPath)
+                           << ": " << errorMessage;
             }
         } // Windows
         if (matches)
@@ -281,7 +281,7 @@ bool runProcess(const QString &binary, const QStringList &args,
     foreach (const QString &a, args)
         appendToCommandLine(a, &commandLine);
     if (optVerboseLevel > 1)
-        std::printf("Running: %s\n", qPrintable(commandLine));
+        std::wcout << "Running: " << commandLine << '\n';
 
     QScopedArrayPointer<wchar_t> commandLineW(new wchar_t[commandLine.size() + 1]);
     commandLine.toWCharArray(commandLineW.data());
@@ -395,7 +395,7 @@ bool runProcess(const QString &binary, const QStringList &args,
         }
 
         if (!workingDirectory.isEmpty() && !QDir::setCurrent(workingDirectory)) {
-            std::fprintf(stderr, "Failed to change working directory to %s.\n", qPrintable(workingDirectory));
+            std::wcerr << "Failed to change working directory to " << workingDirectory << ".\n";
             ::_exit(-1);
         }
 
@@ -510,12 +510,12 @@ QString queryQMake(const QString &variable, QString *errorMessage)
 
 // Update a file or directory.
 bool updateFile(const QString &sourceFileName, const QStringList &nameFilters,
-                const QString &targetDirectory, JsonOutput *json, QString *errorMessage)
+                const QString &targetDirectory, unsigned flags, JsonOutput *json, QString *errorMessage)
 {
     const QFileInfo sourceFileInfo(sourceFileName);
     const QString targetFileName = targetDirectory + QLatin1Char('/') + sourceFileInfo.fileName();
     if (optVerboseLevel > 1)
-        std::printf("Checking %s, %s\n", qPrintable(sourceFileName), qPrintable(targetFileName));
+        std::wcout << "Checking " << sourceFileName << ", " << targetFileName<< '\n';
 
     if (!sourceFileInfo.exists()) {
         *errorMessage = QString::fromLatin1("%1 does not exist.").arg(QDir::toNativeSeparators(sourceFileName));
@@ -540,8 +540,8 @@ bool updateFile(const QString &sourceFileName, const QStringList &nameFilters,
         } else { // exists.
             QDir d(targetDirectory);
             if (optVerboseLevel)
-                std::printf("Creating %s.\n", qPrintable( QDir::toNativeSeparators(targetFileName)));
-            if (!d.mkdir(sourceFileInfo.fileName())) {
+                std::wcout << "Creating " << QDir::toNativeSeparators(targetFileName) << ".\n";
+            if (!(flags & SkipUpdateFile) && !d.mkdir(sourceFileInfo.fileName())) {
                 *errorMessage = QString::fromLatin1("Cannot create directory %1 under %2.")
                                 .arg(sourceFileInfo.fileName(), QDir::toNativeSeparators(targetDirectory));
                 return false;
@@ -551,21 +551,22 @@ bool updateFile(const QString &sourceFileName, const QStringList &nameFilters,
         QDir dir(sourceFileName);
         const QStringList allEntries = dir.entryList(nameFilters, QDir::Files) + dir.entryList(QDir::Dirs | QDir::NoDotAndDotDot);
         foreach (const QString &entry, allEntries)
-            if (!updateFile(sourceFileName + QLatin1Char('/') + entry, nameFilters, targetFileName, json, errorMessage))
+            if (!updateFile(sourceFileName + QLatin1Char('/') + entry, nameFilters, targetFileName, flags, json, errorMessage))
                 return false;
         return true;
     } // Source is directory.
 
     if (targetFileInfo.exists()) {
-        if (targetFileInfo.lastModified() >= sourceFileInfo.lastModified()) {
+        if (!(flags & ForceUpdateFile)
+            && targetFileInfo.lastModified() >= sourceFileInfo.lastModified()) {
             if (optVerboseLevel)
-                std::printf("%s is up to date.\n", qPrintable(sourceFileInfo.fileName()));
+                std::wcout << sourceFileInfo.fileName() << " is up to date.\n";
             if (json)
                 json->addFile(sourceFileName, targetDirectory);
             return true;
         }
         QFile targetFile(targetFileName);
-        if (!targetFile.remove()) {
+        if (!(flags & SkipUpdateFile) && !targetFile.remove()) {
             *errorMessage = QString::fromLatin1("Cannot remove existing file %1: %2")
                             .arg(QDir::toNativeSeparators(targetFileName), targetFile.errorString());
             return false;
@@ -573,8 +574,8 @@ bool updateFile(const QString &sourceFileName, const QStringList &nameFilters,
     } // target exists
     QFile file(sourceFileName);
     if (optVerboseLevel)
-        std::printf("Updating %s.\n", qPrintable(sourceFileInfo.fileName()));
-    if (!file.copy(targetFileName)) {
+        std::wcout << "Updating " << sourceFileInfo.fileName() << ".\n";
+    if (!(flags & SkipUpdateFile) && !file.copy(targetFileName)) {
         *errorMessage = QString::fromLatin1("Cannot copy %1 to %2: %3")
                 .arg(QDir::toNativeSeparators(sourceFileName),
                      QDir::toNativeSeparators(targetFileName),
@@ -619,6 +620,11 @@ bool readElfExecutable(const QString &elfExecutableFileName, QString *errorMessa
 
 #ifdef Q_OS_WIN
 
+static inline QString stringFromRvaPtr(const void *rvaPtr)
+{
+    return QString::fromLocal8Bit((const char *)rvaPtr);
+}
+
 // Helper for reading out PE executable files: Find a section header for an RVA
 // (IMAGE_NT_HEADERS64, IMAGE_NT_HEADERS32).
 template <class ImageNtHeader>
@@ -710,15 +716,24 @@ inline QStringList readImportSections(const ImageNtHeader *ntHeaders, const void
     }
     QStringList result;
     for ( ; importDesc->Name; ++importDesc)
-        result.push_back(QString::fromLocal8Bit((const char *)rvaToPtr(importDesc->Name, ntHeaders, base)));
+        result.push_back(stringFromRvaPtr(rvaToPtr(importDesc->Name, ntHeaders, base)));
+
+    // Read delay-loaded DLLs, see http://msdn.microsoft.com/en-us/magazine/cc301808.aspx .
+    // Check on grAttr bit 1 whether this is the format using RVA's > VS 6
+    if (const DWORD delayedImportsStartRVA = ntHeaders->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DELAY_IMPORT].VirtualAddress) {
+        const ImgDelayDescr *delayedImportDesc = (const ImgDelayDescr *)rvaToPtr(delayedImportsStartRVA, ntHeaders, base);
+        for ( ; delayedImportDesc->rvaDLLName && (delayedImportDesc->grAttrs & 1); ++delayedImportDesc)
+            result.push_back(stringFromRvaPtr(rvaToPtr(delayedImportDesc->rvaDLLName, ntHeaders, base)));
+    }
+
     return result;
 }
 
 // Read a PE executable and determine dependent libraries, word size
-// and debug flags. Note that the debug flag cannot be relied on for MinGW.
+// and debug flags.
 bool readPeExecutable(const QString &peExecutableFileName, QString *errorMessage,
                       QStringList *dependentLibrariesIn, unsigned *wordSizeIn,
-                      bool *isDebugIn)
+                      bool *isDebugIn, bool isMinGW)
 {
     bool result = false;
     HANDLE hFile = NULL;
@@ -763,12 +778,27 @@ bool readPeExecutable(const QString &peExecutableFileName, QString *errorMessage
         bool debug = false;
         if (wordSize == 32) {
             const IMAGE_NT_HEADERS32 *ntHeaders32 = reinterpret_cast<const IMAGE_NT_HEADERS32 *>(ntHeaders);
-            debug = ntHeaders32->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size;
+
+            if (!isMinGW) {
+                debug = ntHeaders32->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size;
+            } else {
+                // Use logic that's used e.g. in objdump / pfd library
+                debug = !(ntHeaders32->FileHeader.Characteristics & IMAGE_FILE_DEBUG_STRIPPED);
+            }
+
             if (dependentLibrariesIn)
                 *dependentLibrariesIn = readImportSections(ntHeaders32, fileMemory, errorMessage);
+
         } else {
             const IMAGE_NT_HEADERS64 *ntHeaders64 = reinterpret_cast<const IMAGE_NT_HEADERS64 *>(ntHeaders);
-            debug = ntHeaders64->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size;
+
+            if (!isMinGW) {
+                debug = ntHeaders64->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_DEBUG].Size;
+            } else {
+                // Use logic that's used e.g. in objdump / pfd library
+                debug = !(ntHeaders64->FileHeader.Characteristics & IMAGE_FILE_DEBUG_STRIPPED);
+            }
+
             if (dependentLibrariesIn)
                 *dependentLibrariesIn = readImportSections(ntHeaders64, fileMemory, errorMessage);
         }
@@ -777,8 +807,8 @@ bool readPeExecutable(const QString &peExecutableFileName, QString *errorMessage
             *isDebugIn = debug;
         result = true;
         if (optVerboseLevel > 1)
-            std::printf("%s: %s %u bit, debug: %d\n", __FUNCTION__,
-                        qPrintable(peExecutableFileName), wordSize, debug);
+            std::wcout << __FUNCTION__ << ": " << peExecutableFileName
+                       << ' ' << wordSize << " bit, debug: " << debug << '\n';
     } while (false);
 
     if (fileMemory)
@@ -793,7 +823,7 @@ bool readPeExecutable(const QString &peExecutableFileName, QString *errorMessage
     return result;
 }
 
-QString findD3dCompiler(Platform platform, unsigned wordSize)
+QString findD3dCompiler(Platform platform, const QString &qtBinDir, unsigned wordSize)
 {
     const QString prefix = QStringLiteral("D3Dcompiler_");
     const QString suffix = QLatin1String(windowsSharedLibrarySuffix);
@@ -813,12 +843,27 @@ QString findD3dCompiler(Platform platform, unsigned wordSize)
                 return files.front().absoluteFilePath();
         }
     }
+    QStringList candidateVersions;
+    for (int i = 47 ; i >= 40 ; --i)
+        candidateVersions.append(prefix + QString::number(i) + suffix);
+    // Check the bin directory of the Qt SDK (in case it is shadowed by the
+    // Windows system directory in PATH).
+    foreach (const QString &candidate, candidateVersions) {
+        const QFileInfo fi(qtBinDir + QLatin1Char('/') + candidate);
+        if (fi.isFile())
+            return fi.absoluteFilePath();
+    }
     // Find the latest D3D compiler DLL in path (Windows 8.1 has d3dcompiler_47).
     if (platform & IntelBased) {
-        for (int i = 47 ; i >= 40 ; --i) {
-            const QString dll = findInPath(prefix + QString::number(i) + suffix);
-            if (!dll.isEmpty())
+        QString errorMessage;
+        unsigned detectedWordSize;
+        foreach (const QString &candidate, candidateVersions) {
+            const QString dll = findInPath(candidate);
+            if (!dll.isEmpty()
+                && readPeExecutable(dll, &errorMessage, 0, &detectedWordSize, 0)
+                && detectedWordSize == wordSize) {
                 return dll;
+            }
         }
     }
     return QString();
@@ -833,7 +878,7 @@ bool readPeExecutable(const QString &, QString *errorMessage,
     return false;
 }
 
-QString findD3dCompiler(Platform, unsigned)
+QString findD3dCompiler(Platform, const QString &, unsigned)
 {
     return QString();
 }
